/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2023 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#ifndef OSM_DATA_H
#define OSM_DATA_H

#include <librnd/core/global_typedefs.h>
#include "osm_typedefs.h"
#include <librnd/core/hidlib.h>
#include <librnd/core/color.h>
#include <librnd/font/xform_mx.h>
#include <libuundo/uundo.h>
#include <genvector/vtp0.h>
#include "rtree.h"
#include "htIp.h"

#define OSM_ZORD_MAX 16
#define OSM_REFANG_MAX 8


struct osm_design_s {
	rnd_design_t hidlib; /* shall be the first */

	/* UI states */
	rnd_coord_t crosshair_x, crosshair_y;
	osm_node_t *crosshair_snap_node;
	double *refang;
	double refangs[OSM_REFANG_MAX];

	/* data */
	vtp0_t layers; /* to osm_layer_t */
	int lysel;     /* layer idx (lid) of the currently selected layer */
	osm_rtree_box_t bbox;

	/* for loaders */
	struct {
		long ly;
	} loader;
};

struct osm_layer_s {
	/* all objects */
	osm_rtree_t objs;

	/* objects by display levels (z order); 0 is bottom (rendered first) */
	osm_rtree_t zord[OSM_ZORD_MAX];

	htIp_t id2obj[OSM_OBJ_max]; /* the same id is reused for a node, a way and a relation */
	char *name, *short_name;
	osm_design_t *parent;
	long rev; /* increased by 1 when objects are added or removed */
	unsigned vis:1;
	unsigned editable:1;  /* if 0, the layer is imagery or gpx or something other that's not for the database */
	const char *force_style_name; /* if not NULL: pick style based on this name because of the layer type */

	/* for plugin-rendered layers */
	void (*plugin_draw)(osm_layer_t *ly, rnd_hid_gc_t gc, const rnd_hid_expose_ctx_t *region, rnd_xform_t *xform_caller);
	void (*plugin_set_offs)(osm_layer_t *ly, rnd_coord_t ox, rnd_coord_t oy, int temporary);
	void (*plugin_get_offs)(osm_layer_t *ly, rnd_coord_t *ox, rnd_coord_t *oy);
	void *plugin_data;

	uundo_list_t undo;
	osm_id_t last_new_id; /* last id assigned to a new object; negative, starts at 0; decrease and use */
	vtp0_t remove_list;   /* non-blank removed objects for undoable remove; may contain NULL elements */
};

extern osm_design_t osm;

void osm_layer_init(osm_layer_t *layer);
osm_layer_t *osm_layer_new(void);

/* Create layer by unique name, return existing layer on name match, creates
   non-existing layer if alloc is non-zero; appends to design automatically.
   The bottom variant allocates new layer on the bottom so it's rendered under
   everything else. */
osm_layer_t *osm_layer_by_name(osm_design_t *osm, const char *name, int alloc);
osm_layer_t *osm_layer_by_name_bottom(osm_design_t *osm, const char *name, int alloc);

/* Returns index of the new layer */
long osm_layer_append_to_design(osm_design_t *osm, osm_layer_t *layer, int bottom);

/* Invent a layer color for a main layer that has non set explicitly */
void osm_layer_invent_color(osm_design_t *osm, osm_layer_t *layer);

/* If layer is part of a design, remove it. Destroy all layer objects
   and free all memory allocated by the objects or the layer. */
void osm_layer_free_fields(osm_layer_t *ly);

/* Free all fields of layer and layer itself as well */
void osm_layer_destroy(osm_layer_t *layer);

/* Return currently selected layer or NULL */
osm_layer_t *osm_layer_get_curr(osm_design_t *osm);

int osm_layer_set_vis(osm_design_t *osm, rnd_cardinal_t lid, int vis, int emit_event);

void osm_design_free_fields(osm_design_t *osm);

int osm_is_empty(osm_design_t *osm);

void osm_data_bbox(osm_design_t *osm);

/* Change layer selection. -1 means none selected */
void osm_layer_select(osm_design_t *osm, int idx);

void osm_data_reverse_layers(osm_design_t *osm, int inhibit_ev);

#define osm_hid_redraw(osm) rnd_render->invalidate_all(rnd_render)

extern char osm_measurement_layer_name[];

osm_id_t osm_str2id(const char *str, char **end);

void osm_layer_clear_remove_list(osm_layer_t *ly);
void osm_layer_clear_objs(osm_layer_t *ly); /* remove and free all ojects */


#endif
