/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

/* included from obj_relation.c to split the long code */

#include <librnd/core/rnd_printf.h>
#include "undo.h"

static const char obj_relation_cookie[] = "relation";

typedef struct {
	osm_id_t id;
	osm_layer_t *ly;
} undo_relation_fork_t;

static int undo_relation_fork_swap(void *udata)
{
	undo_relation_fork_t *u = udata;
	osm_relation_t *relation = htIp_get(&u->ly->id2obj[OSM_OBJ_RELATION], u->id);

	if (relation == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_relation_fork_swap(): relation not found: " OSM_PID "\n", u->id);
		return -1;
	}

	if (relation->curr == &relation->orig)
		osm_relation_fork(relation);
	else
		osm_relation_unfork(relation);

	return 0;
}

static void undo_relation_fork_print(void *udata, char *dst, size_t dst_len)
{
	undo_relation_fork_t *u = udata;
	osm_relation_t *relation = htIp_get(&u->ly->id2obj[OSM_OBJ_RELATION], u->id);

	if (relation == NULL) {
		rnd_snprintf(dst, dst_len, "relation fork: invalid id " OSM_PID, u->id);
		return;
	}

	rnd_snprintf(dst, dst_len, "relation %s", (relation->curr == &relation->orig) ? "fork" : "unfork");
}


static const uundo_oper_t undo_relation_fork = {
	obj_relation_cookie,
	NULL,
	undo_relation_fork_swap,
	undo_relation_fork_swap,
	undo_relation_fork_print
};


void osm_relation_modify_fork(osm_layer_t *layer, osm_relation_t *relation, int undoable)
{
	undo_relation_fork_t utmp, *u = &utmp;

	if (undoable) u = uundo_append(&layer->undo, &undo_relation_fork, sizeof(undo_relation_fork_t));

	u->ly = layer;
	u->id = relation->hdr.id;

	undo_relation_fork_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
}




typedef struct {
	osm_id_t id;
	osm_layer_t *ly;

	osm_relmember_t member;

	long member_idx;
	enum {
		WMOD_ADD,
		WMOD_DEL,
		WMOD_REPLACE
	} op;
} undo_relation_modify_t;

static int undo_relation_modify_swap(void *udata)
{
	undo_relation_modify_t *u = udata;
	osm_relation_t *relation = htIp_get(&u->ly->id2obj[OSM_OBJ_RELATION], u->id);

	if (relation == NULL) {
		rnd_message(RND_MSG_ERROR, "undo_relation_modify_swap(): relation not found: " OSM_PID "\n", u->id);
		return -1;
	}

	TODO("Figure redraw");
/* osm_obj_redraw(relation); */


	osm_relation_pre_edit((osm_any_obj_t *)relation, 1, 1);

	switch(u->op) {
		case WMOD_ADD:
			assert(u->member_idx >= 0);
			assert(u->member_idx <= relation->curr->members.used);
			vtr0_insert_len(&relation->curr->members, u->member_idx, &u->member, 1);
			memset(&u->member, 0, sizeof(u->member));
			u->op = WMOD_DEL;
			break;

		case WMOD_DEL:
			assert(u->member_idx >= 0);
			assert(u->member_idx < relation->curr->members.used);
			memcpy(&u->member, &relation->curr->members.array[u->member_idx], sizeof(u->member));
			vtr0_remove(&relation->curr->members, u->member_idx, 1);
			u->op = WMOD_ADD;
			break;

		case WMOD_REPLACE:
			assert(u->member_idx >= 0);
			assert(u->member_idx <= relation->curr->members.used);
			rnd_swap(osm_relmember_t, relation->curr->members.array[u->member_idx], u->member);
			break;
	}

	osm_relation_post_edit((osm_any_obj_t *)relation, 0);

	TODO("Figure layer changed flag");
	TODO("Figure redraw");
/* osm_obj_redraw(relation); */

	return 0;
}

static void undo_relation_modify_print(void *udata, char *dst, size_t dst_len)
{
	undo_relation_modify_t *u = udata;
	osm_relation_t *relation = htIp_get(&u->ly->id2obj[OSM_OBJ_RELATION], u->id);
	char *opname;

	if (relation == NULL) {
		rnd_snprintf(dst, dst_len, "relation modify: invalid id " OSM_PID, u->id);
		return;
	}

	switch(u->op) {
		case WMOD_ADD:     opname = "add"; break;
		case WMOD_DEL:     opname = "del"; break;
		case WMOD_REPLACE: opname = "replace"; break;
	}

	rnd_snprintf(dst, dst_len, "relation #%ld %s member {%d:#" OSM_PID ":%s} at %ld",
		u->id, opname, u->member.type, u->member.ref, u->member.role, u->member_idx);
}

static void undo_relation_modify_free(void *udata)
{
	undo_relation_modify_t *u = udata;
	free(u->member.role);
	u->member.role = NULL;
}

static const uundo_oper_t undo_relation_modify = {
	obj_relation_cookie,
	undo_relation_modify_free,
	undo_relation_modify_swap,
	undo_relation_modify_swap,
	undo_relation_modify_print
};


void osm_relation_modify_append(osm_layer_t *layer, osm_relation_t *relation, osm_objtype_t typ, osm_id_t id, const char *role, int undoable)
{
	undo_relation_modify_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (relation->curr == &relation->orig)
			osm_relation_modify_fork(layer, relation, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_relation_modify, sizeof(undo_relation_modify_t));

	u->ly = layer;
	u->id = relation->hdr.id;
	u->member.type = typ;
	u->member.ref = id;
	u->member.role = rnd_strdup(role);
	u->member_idx = relation->curr->members.used;
	u->op = WMOD_ADD;

	undo_relation_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_relation_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

void osm_relation_modify_insert_before(osm_layer_t *layer, osm_relation_t *relation, osm_id_t before_member_idx, osm_objtype_t typ, osm_id_t id, const char *role, int undoable)
{
	undo_relation_modify_t utmp, *u = &utmp;

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (relation->curr == &relation->orig)
			osm_relation_modify_fork(layer, relation, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_relation_modify, sizeof(undo_relation_modify_t));

	u->ly = layer;
	u->id = relation->hdr.id;
	u->member.type = typ;
	u->member.ref = id;
	u->member.role = rnd_strdup(role);
	u->member_idx = before_member_idx;
	u->op = WMOD_ADD;

	undo_relation_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_relation_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}


void osm_relation_modify_remove(osm_relation_t *relation, long idx, int undoable)
{
	undo_relation_modify_t utmp, *u = &utmp;
	osm_layer_t *layer = relation->hdr.layer;

	assert(idx >= 0);
	assert(idx < relation->curr->members.used);

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (relation->curr == &relation->orig)
			osm_relation_modify_fork(layer, relation, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_relation_modify, sizeof(undo_relation_modify_t));

	u->ly = layer;
	u->id = relation->hdr.id;
	u->member_idx = idx;
	u->op = WMOD_DEL;

	undo_relation_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_relation_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}

void osm_relation_modify_member(osm_relation_t *relation, long idx, osm_objtype_t *typ, osm_id_t *id, const char **role, int undoable)
{
	undo_relation_modify_t utmp, *u = &utmp;
	osm_layer_t *layer = relation->hdr.layer;
	osm_relmember_t *m;

	assert(idx >= 0);
	assert(idx < relation->curr->members.used);

	m = &relation->curr->members.array[idx];

	if (undoable) {
		uundo_freeze_serial(&layer->undo);

		if (relation->curr == &relation->orig)
			osm_relation_modify_fork(layer, relation, undoable);
	}

	if (undoable) u = uundo_append(&layer->undo, &undo_relation_modify, sizeof(undo_relation_modify_t));

	u->ly = layer;
	u->id = relation->hdr.id;

	u->member.type = (typ == NULL) ? m->type : *typ;
	u->member.ref  = (id == NULL) ? m->ref : *id;
	u->member.role = rnd_strdup((id == NULL) ? m->role : *role);

	u->member_idx = idx;
	u->op = WMOD_REPLACE;

	undo_relation_modify_swap(u);
	if (undoable) osm_undo_inc_serial(layer);
	else undo_relation_modify_free(u);

	if (undoable) {
		uundo_unfreeze_serial(&layer->undo);
		uundo_inc_serial(&layer->undo);
	}
}
