/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - server communication
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */


#include "config.h"
#include "obj_any.h"

#define PID "\"" OSM_PID "\""

RND_INLINE long osmchange_gen_xml_obj_(FILE *f, osm_any_obj_t *obj, osm_id_t changeset_id)
{
	osm_oxml_fprint_obj(f, "   ", obj, &changeset_id);
	return 1;
}

RND_INLINE int obj_changed(osm_any_obj_t *obj)
{
	if (obj->hdr.id < 0)
		return 1; /* new object */

	switch(obj->hdr.type) {
		case OSM_OBJ_NODE:      return (obj->node.curr != &obj->node.orig);
		case OSM_OBJ_WAY:       return (obj->way.curr != &obj->way.orig);
		case OSM_OBJ_RELATION:  return (obj->relation.curr != &obj->relation.orig);

		case OSM_OBJ_CHANGESET:
		case OSM_OBJ_invalid: case OSM_OBJ_vmax: abort();
	}

	/* can't get here */
	abort();
	return -1;
}


/* Write a single object if changed. Return number of object printed or -1 on error. */
RND_INLINE int osmchange_gen_xml_obj(FILE *f, osm_any_obj_t *obj, osm_id_t changeset_id, int force_del)
{
	long res;

	if ((obj->hdr.type < 0) || (obj->hdr.type >= OSM_OBJ_max)) {
		rnd_message(RND_MSG_ERROR, "osmchange_gen_xml_obj(): invalid object type %d\n", (int)obj->hdr.type);
		return -1;
	}

	if (!obj_changed(obj) && !force_del)
		return 0;

	if (!obj->hdr.visible || force_del) {
		if (obj->hdr.id < 0) /* don't commit removal of newly created objects */
			return 0;
		fprintf(f, " <delete>\n");
		fprintf(f, "  <%s id="PID" changeset="PID" version=\"%ld\"/>\n", osm_obj_name[obj->hdr.type], obj->hdr.id, changeset_id, obj->hdr.version);
		fprintf(f, " </delete>\n");
		res = 1;
	}
	else if (obj->hdr.id < 0) {
		fprintf(f, " <create>\n");
		res = osmchange_gen_xml_obj_(f, obj, changeset_id);
		fprintf(f, " </create>\n");
	}
	else {
		fprintf(f, " <modify>\n");
		res = osmchange_gen_xml_obj_(f, obj, changeset_id);
		fprintf(f, " </modify>\n");
	}

	return res;
}


/* Write all objects matching type within area. Return number of object printed or -1 on error. */
RND_INLINE long osmchange_gen_xml_objs(FILE *f, osm_layer_t *ly, osm_rtree_box_t *area, osm_id_t changeset_id, osm_objtype_t type)
{
	osm_rtree_it_t it;
	void *o;
	long r, res = 0;

	if (area == NULL) {
		for(o = osm_rtree_all_first(&it, &ly->objs); o != NULL; o = osm_rtree_all_next(&it)) {
			osm_any_obj_t *obj = o;
			if (obj->hdr.type == type) {
				r = osmchange_gen_xml_obj(f, obj, changeset_id, 0);
				if (r < 0) return -1;
				res += r;
			}
		}
	}
	else {
		for(o = osm_rtree_first(&it, &ly->objs, area); o != NULL; o = osm_rtree_next(&it)) {
			osm_any_obj_t *obj = o;
			if (obj->hdr.type == type) {
				r = osmchange_gen_xml_obj(f, obj, changeset_id, 0);
				if (r < 0) return -1;
				res += r;
			}
		}
	}

	return res;
}

RND_INLINE long osmchange_gen_xml_dels(FILE *f, osm_layer_t *ly, osm_rtree_box_t *area, osm_id_t changeset_id, osm_objtype_t type)
{
	long n, r, res = 0;

	for(n = 0; n < ly->remove_list.used; n++) {
		osm_any_obj_t *obj = ly->remove_list.array[n];
		if ((obj != NULL) && (obj->hdr.type == type) && (obj->hdr.id > 0)) {
			r = osmchange_gen_xml_obj(f, obj, changeset_id, 1);
			if (r < 0) return -1;
			res += r;
		}
	}

	return res;
}

/* Write all layer objects within area. Return number of object printed or -1 on error. */
int osmchange_gen_xml(FILE *f, osm_layer_t *ly, osm_rtree_box_t *area, osm_id_t changeset_id)
{
	long r1, r2, r3, r4, r5, r6;

	fprintf(f, "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
	fprintf(f, "<osmChange version=\"0.6\" generator=\"osm-rnd " OSM_VERS " osmchange_gen_xml()\">\n");

	/* generate per object type in order of how objects can reference to other
	   objects. This helps avoiding forward references. */
	r1 = osmchange_gen_xml_objs(f, ly, area, changeset_id, OSM_OBJ_NODE);
	r2 = osmchange_gen_xml_objs(f, ly, area, changeset_id, OSM_OBJ_WAY);
	r3 = osmchange_gen_xml_objs(f, ly, area, changeset_id, OSM_OBJ_RELATION);

	r4 = osmchange_gen_xml_dels(f, ly, area, changeset_id, OSM_OBJ_RELATION);
	r5 = osmchange_gen_xml_dels(f, ly, area, changeset_id, OSM_OBJ_WAY);
	r6 = osmchange_gen_xml_dels(f, ly, area, changeset_id, OSM_OBJ_NODE);

	fprintf(f, "</osmChange>\n");

	if ((r1 < 0) || (r2 < 0) || (r3 < 0) || (r4 < 0) || (r5 < 0) || (r6 < 0))
		return -1;

	return r1 + r2 + r3 + r4 + r5 + r6;
}

