/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Presets for tagging
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include <librnd/hid/hid_dad.h>
#include <librnd/hid/hid_dad_tree.h>
#include <librnd/src_3rd/genregex/regex_sei.h>

#include "edit_act.h"
#include "funchash_core.h"

typedef struct{
	RND_DAD_DECL_NOINIT(dlg)
	int wlist, wfilt, wtags;

	lht_node_t *clicked;

	/* object address if not selection */
	unsigned selection:1;
	osm_design_t *osm;
	osm_layer_t *ly;
	osm_objtype_t obj_type;
	osm_id_t obj_id;
} presets_ctx_t;

static void presets_close_cb(void *caller_data, rnd_hid_attr_ev_t ev)
{
	presets_ctx_t *ctx = caller_data;

	RND_DAD_FREE(ctx->dlg);
	free(ctx);
}


static void presets_db2dlg(presets_ctx_t *ctx)
{
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	rnd_hid_row_t *r;
	char *cell[2], *cursor_path = NULL;
	rnd_hid_attr_val_t hv;
	long n;

	/* reload the list */
	attr = &ctx->dlg[ctx->wlist];
	tree = attr->wdata;

	/* remember cursor */
	r = rnd_dad_tree_get_selected(attr);
	if (r != NULL)
		cursor_path = rnd_strdup(r->cell[0]);

	/* remove existing items */
	rnd_dad_tree_clear(tree);

	cell[1] = NULL;

	for(n = 0; n < presets_sorted.used; n++) {
		lht_node_t *nd = presets_sorted.array[n];
		cell[0] = rnd_strdup(nd->name);
		rnd_dad_tree_append(attr, NULL, cell);
	}

	/* restore cursor */
	if (cursor_path != NULL) {
		hv.str = cursor_path;
		rnd_gui->attr_dlg_set_value(ctx->dlg_hid_ctx, ctx->wlist, &hv);
	}

	free(cursor_path);
}

static rnd_hid_row_t *prst_tag_add(presets_ctx_t *ctx, const char *key, const char *val)
{
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wtags];
	rnd_hid_tree_t *tree = attr->wdata;
	rnd_hid_row_t *r, *found = NULL;

	for(r = gdl_first(&tree->rows); r != NULL; r = gdl_next(&tree->rows, r)) {
		if (strcmp(r->cell[0], key) == 0) {
			found = r;
			break;
		}
	}

	if (found == NULL) {
		char *cell[3];
		cell[0] = rnd_strdup(key);
		cell[1] = rnd_strdup(val == NULL ? "" : val);
		cell[2] = NULL;
		return rnd_dad_tree_append(attr, NULL, cell);
	}

	if (val != NULL)
		rnd_dad_tree_modify_cell(attr, found, 1, rnd_strdup(val));
	return found;
}

static void prst_activate(presets_ctx_t *ctx)
{
	lht_node_t *nd = ctx->clicked, *tags, *all_fields, *my_fields, *field, *n, *key, *k;
	lht_dom_iterator_t it, it2;
	rnd_hid_row_t *r;

	/* copy tags */
	tags = lht_dom_hash_get(nd, "tags");
	if (tags != NULL) {
		for(n = lht_dom_first(&it, tags); n != NULL; n = lht_dom_next(&it))
			if (n->type == LHT_TEXT)
				prst_tag_add(ctx, n->name, n->data.text.value);
	}

	/* copy fields */
	all_fields = lht_dom_hash_get(presets->root, "fields");
	my_fields = lht_dom_hash_get(nd, "fields");
	if ((all_fields != NULL) && (my_fields != NULL)) {
		for(n = lht_dom_first(&it, my_fields); n != NULL; n = lht_dom_next(&it)) {
			if (n->type != LHT_TEXT) continue;
			field = lht_dom_hash_get(all_fields, n->data.text.value);
			if (field != NULL) {
				key = lht_dom_hash_get(field, "keys");
				if (key != NULL) { /* process a list of keys */
					for(k = lht_dom_first(&it2, key); k != NULL; k = lht_dom_next(&it2)) {
						if (k->type == LHT_TEXT) {
							r = prst_tag_add(ctx, k->data.text.value, NULL);
							r->user_data = k;
						}
						else
							rnd_message(RND_MSG_ERROR, "Presets database error: 'keys' for field '%s' contains non-text nodes\n", n->data.text.value);
					}
				}
				else { /* process a single key */
					key = lht_dom_hash_get(field, "key");
					if ((key != NULL) && (key->type == LHT_TEXT)) {
						r = prst_tag_add(ctx, key->data.text.value, NULL);
						r->user_data = key;
					}
				}
			}
			else
				rnd_message(RND_MSG_ERROR, "preset: field '%s' not found\n", n->name);
		}
	}
}

static void prst_unhide(rnd_hid_tree_t *tree, gdl_list_t *rowlist, re_sei_t *preg)
{
	rnd_hid_row_t *r, *pr;

	for(r = gdl_first(rowlist); r != NULL; r = gdl_next(rowlist, r)) {
		if ((preg == NULL) || (re_sei_exec(preg, r->cell[0]))) {
			rnd_dad_tree_hide_all(tree, &r->children, 0); /* if this is a node with children, show all children */
			for(pr = r; pr != NULL; pr = rnd_dad_tree_parent_row(tree, pr)) /* also show all parents so it is visible */
				pr->hide = 0;
		}
		prst_unhide(tree, &r->children, preg);
	}
}

static void prst_filter_cb(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_inp__)
{
	presets_ctx_t *ctx = caller_data;
	rnd_hid_attribute_t *attr_inp = &ctx->dlg[ctx->wfilt];
	rnd_hid_attribute_t *attr;
	rnd_hid_tree_t *tree;
	const char *otext;
	char *text;
	int have_filter_text;

	attr = &ctx->dlg[ctx->wlist];
	tree = attr->wdata;
	otext = attr_inp->val.str;
	if (otext == NULL) otext = "";
	text = rnd_strdup(otext);
	have_filter_text = (*text != '\0');

	/* hide or unhide everything */

	if (have_filter_text) {
		/* need to unhide for expand to work */
		rnd_dad_tree_hide_all(tree, &tree->rows, 0);
		rnd_dad_tree_update_hide(attr);
		rnd_dad_tree_expcoll(attr, NULL, 1, 1);
		rnd_dad_tree_hide_all(tree, &tree->rows, 1);
	}
	else
		rnd_dad_tree_hide_all(tree, &tree->rows, 0);

	if (have_filter_text) { /* unhide hits and all their parents */
		re_sei_t *regex = NULL;

		if ((text != NULL) && (*text != '\0'))
			regex = re_sei_comp(text);

		prst_unhide(tree, &tree->rows, regex);

		if (regex != NULL)
			re_sei_free(regex);
	}

	rnd_dad_tree_update_hide(attr);

	free(text);
}

static void prst_select(rnd_hid_attribute_t *attrib, void *hid_ctx, rnd_hid_row_t *row)
{
	rnd_hid_tree_t *tree = attrib->wdata;
	presets_ctx_t *ctx = tree->user_ctx;
	lht_node_t *parent, *nd;

	parent = lht_dom_hash_get(presets->root, "presets");
	if (parent == NULL)
		return;

	nd = lht_dom_hash_get(parent, row->cell[0]);
	if (nd == NULL)
		return;

	if (ctx->clicked == nd) {
		prst_activate(ctx);
	}
	else {
		ctx->clicked = nd;
	}
}

#include "dlg_field.c"

static void cb_edit_field(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_btn)
{
	presets_ctx_t *ctx = caller_data;
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wtags];
	rnd_hid_row_t *r = rnd_dad_tree_get_selected(attr);
	lht_node_t *field;

	if (r == NULL) {
		rnd_message(RND_MSG_ERROR, "Select a tag first\n");
		return;
	}

	field = r->user_data;
	if (field == NULL) {
		rnd_message(RND_MSG_ERROR, "There is no preset field available for this tag\n");
		return;
	}

	/* field is a node in the main ha:field subtree */
	presets_field_dlg(ctx, r, field);
}

#define MAX_TAG_ARGS 1024
static void apply_print(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_btn, int print)
{
	presets_ctx_t *ctx = caller_data;
	int argc;
	fgw_arg_t args[MAX_TAG_ARGS], res;
	rnd_hid_attribute_t *attr = &ctx->dlg[ctx->wtags];
	rnd_hid_tree_t *tree = attr->wdata;
	rnd_hid_row_t *r = NULL;

	args[1].type = FGW_STR; args[1].val.cstr = "selected";
	argc = 2;

	if (print)
		rnd_message(RND_MSG_INFO, "tagset(selected");

	for(r = gdl_first(&tree->rows); r != NULL; r = gdl_next(&tree->rows, r)) {
		if (argc >= (MAX_TAG_ARGS-1)) {
			rnd_message(RND_MSG_ERROR, "Preset error: too many tags - some of them are not set\n");
			break;
		}
		if ((r->cell[1] == NULL) || (*r->cell[1] == '\0'))
			continue;
		if (!print) {
			args[argc].type = FGW_STR; args[argc].val.cstr = r->cell[0];
			argc++;
			args[argc].type = FGW_STR; args[argc].val.cstr = r->cell[1];
			argc++;
		}
		else
			rnd_message(RND_MSG_INFO, ", %s, %s", r->cell[0], r->cell[1]);
	}

	if (print) {
		rnd_message(RND_MSG_INFO, ")\n");
		return;
	}

	if (ctx->selection) {
		rnd_actionv_bin(&ctx->osm->hidlib, "tagset", &res, argc, args);
	}
	else {
		osm_any_obj_t *obj = osm_obj_by_id(ctx->ly, ctx->obj_type, ctx->obj_id);
		if (obj != NULL)
			osm_tagset(obj, argc-2, args+2);
		else
			rnd_message(RND_MSG_ERROR, "Preset error: object not found\n");
	}
}

static void cb_apply(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_btn)
{
	apply_print(hid_ctx, caller_data, attr_btn, 0);
}

static void cb_print(void *hid_ctx, void *caller_data, rnd_hid_attribute_t *attr_btn)
{
	apply_print(hid_ctx, caller_data, attr_btn, 1);
}


static void presets_dlg(osm_design_t *osm, osm_layer_t *ly, osm_any_obj_t *obj, int selection)
{
	presets_ctx_t *ctx;
	const char *title;
	const char *lhdr[] = {"preset name"};
	const char *rhdr[] = {"key", "value"};
	rnd_hid_dad_buttons_t clbtn[] = {{"close", 0}, {NULL, 0}};

	if (presets == NULL)
		presets_load();

	if (presets == NULL) {
		rnd_message(RND_MSG_ERROR, "No presets database loaded\n");
		return;
	}

	ctx = calloc(sizeof(presets_ctx_t), 1);

	ctx->osm = osm;
	ctx->selection = selection;
	ctx->ly = ly;

	if (obj != NULL) {
		ctx->obj_type = obj->hdr.type;
		ctx->obj_id = obj->hdr.id;
	}

	RND_DAD_BEGIN_VBOX(ctx->dlg);
		RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

		RND_DAD_BEGIN_HPANE(ctx->dlg, "left-right");
			RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* left */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_TREE(ctx->dlg, 1, 0, lhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wlist = RND_DAD_CURRENT(ctx->dlg);
					RND_DAD_TREE_SET_CB(ctx->dlg, selected_cb, prst_select);
					RND_DAD_TREE_SET_CB(ctx->dlg, ctx, ctx);

				RND_DAD_BEGIN_HBOX(ctx->dlg); /* left bottom */
					RND_DAD_STRING(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_INIT_FOCUS);
						RND_DAD_HELP(ctx->dlg, "filter: display only presets matching this text\n(if empty: display all)");
						RND_DAD_CHANGE_CB(ctx->dlg, prst_filter_cb);
						ctx->wfilt = RND_DAD_CURRENT(ctx->dlg);
				RND_DAD_END(ctx->dlg);

			RND_DAD_END(ctx->dlg);

			RND_DAD_BEGIN_VBOX(ctx->dlg); /* right */
				RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL);

				RND_DAD_LABEL(ctx->dlg, "<preset description>");

				RND_DAD_TREE(ctx->dlg, 2, 0, rhdr);
					RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL | RND_HATF_SCROLL);
					ctx->wtags = RND_DAD_CURRENT(ctx->dlg);

				RND_DAD_BEGIN_HBOX(ctx->dlg); /* buttons */

					RND_DAD_BUTTON(ctx->dlg, "del");
/*						RND_DAD_CHANGE_CB(ctx->dlg, cb_del);*/
					RND_DAD_BUTTON(ctx->dlg, "new");
/*						RND_DAD_CHANGE_CB(ctx->dlg, cb_new);*/
					RND_DAD_BUTTON(ctx->dlg, "rename");
/*						RND_DAD_CHANGE_CB(ctx->dlg, cb_rename);*/
					RND_DAD_BUTTON(ctx->dlg, "edit field");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_edit_field);

					RND_DAD_BEGIN_VBOX(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL); /* spring */
					RND_DAD_END(ctx->dlg);

					RND_DAD_BUTTON(ctx->dlg, "print");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_print);
					RND_DAD_BUTTON(ctx->dlg, "Apply!");
						RND_DAD_CHANGE_CB(ctx->dlg, cb_apply);

					RND_DAD_BEGIN_VBOX(ctx->dlg);
						RND_DAD_COMPFLAG(ctx->dlg, RND_HATF_EXPFILL); /* spring */
					RND_DAD_END(ctx->dlg);

					RND_DAD_BUTTON_CLOSES(ctx->dlg, clbtn);
				RND_DAD_END(ctx->dlg); /* buttons */
			RND_DAD_END(ctx->dlg); /* left */
		RND_DAD_END(ctx->dlg); /* pane */
	RND_DAD_END(ctx->dlg);

	title = selection ? "Presets on selection" : "Presets on object";

	RND_DAD_NEW("presets", ctx->dlg, title, ctx, 0, presets_close_cb);

	presets_db2dlg(ctx);
}

const char osm_acts_Presets[] = "Presets(selection|object)";
const char osm_acth_Presets[] = "Change tags using presets\n";
fgw_error_t osm_act_Presets(fgw_arg_t *res, int argc, fgw_arg_t *argv)
{
	rnd_design_t *hl = RND_ACT_DESIGN;
	osm_design_t *osm = (osm_design_t *)hl;
	osm_layer_t *ly = osm_layer_get_curr(osm);
	int op;
	osm_any_obj_t *obj;
	rnd_coord_t x, y;

	if ((ly == NULL) || !ly->editable) {
		rnd_message(RND_MSG_ERROR, "Need an editable layer\n");
		return -1;
	}

	RND_ACT_CONVARG(1, FGW_KEYWORD, Presets, op = fgw_keyword(&argv[1]));

	switch(op) {
		case F_Object:
			rnd_hid_get_coords("Presets: click on object", &x, &y, 0);
			obj = osm_search_obj_xy(ly, x, y);
			if (obj == NULL) {
				rnd_message(RND_MSG_ERROR, "No object under the crosshair\n");
				return -1;
			}
			presets_dlg(osm, ly, obj, 0);
			break;
		case F_Selection:
			presets_dlg(osm, ly, NULL, 1);
			break;
	}

	return 0;
}
