/*
 *                            COPYRIGHT
 *
 *  osm-rnd - OpenStreetMap editor/viewer - Optimizations
 *  Copyright (C) 2024 Tibor 'Igor2' Palinkas
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 *  Contact:
 *    Project page: http://repo.hu/projects/osm-rnd
 *    lead developer: http://repo.hu/projects/osm-rnd/contact.html
 *    mailing list: (not yet)
 */

#include "config.h"
#include <math.h>
#include <genlist/gendlist.h>
#include <librnd/config.h>
#include <librnd/core/actions.h>
#include <librnd/core/rotate.h>
#include <librnd/hid/hid.h>

#include "data.h"
#include "event.h"
#include "drc.h"
#include "search.h"
#include "obj_node.h"
#include "obj_way.h"

static const char std_drc_cookie[] = "std_drc";

RND_INLINE void std_drc_node(osm_node_t *node)
{
	osm_any_obj_t *obj = (osm_any_obj_t *)node;

	/* untagged nodes that are not in any relation or way */
	if ((node->curr->tags.used == 0) && !osm_search_has_referer_way(&obj->node) && !osm_search_has_referer_relations(obj))
		osm_drc_error(obj, "Node without tags");
}

RND_INLINE void std_drc_way(osm_way_t *way)
{
	osm_any_obj_t *obj = (osm_any_obj_t *)way;
	long n;

	/* untagged ways that are not in any relation */
	if ((way->curr->tags.used == 0) && !osm_search_has_referer_relations(obj))
		osm_drc_error(obj, "Way without tags");

	if (way->curr->nd.used < 2)
		osm_drc_error(obj, "Way with too few nodes");

	for(n = 1; n < way->curr->nd.used; n++) {
		if (way->curr->nd.array[n-1] == way->curr->nd.array[n]) {
			osm_drc_error(obj, "Duplicate adjacent nodes in way");
			break;
		}
	}
}

RND_INLINE void std_drc_relation(osm_relation_t *rel)
{
}

static void std_drc_object_ev(rnd_design_t *hidlib, void *user_data, int argc, rnd_event_arg_t argv[])
{
	osm_any_obj_t *obj;

	if ((argc != 2) || (argv[1].type != RND_EVARG_PTR))
		return;

	obj = argv[1].d.p;
	switch(obj->hdr.type) {
		case OSM_OBJ_NODE: std_drc_node(&obj->node); break;
		case OSM_OBJ_WAY: std_drc_way(&obj->way); break;
		case OSM_OBJ_RELATION: std_drc_relation(&obj->relation); break;
		case OSM_OBJ_CHANGESET: case OSM_OBJ_vmax: case OSM_OBJ_invalid: break;
	}
}



/*** plugin admin ***/

int pplg_check_ver_std_drc(int ver_needed) { return 0; }

void pplg_uninit_std_drc(void)
{
	rnd_event_unbind_allcookie(std_drc_cookie);
}

int pplg_init_std_drc(void)
{
	rnd_event_bind(OSM_EVENT_DRC_OBJECT, std_drc_object_ev, NULL, std_drc_cookie);
	return 0;
}
